#!/usr/bin/env gosh

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; ---START GAUCHE PRELUDE---

(use srfi-1)
(use srfi-2)
(use srfi-9)

(define define-inline define)

(define-constant arithmetic-shift ash)
(define-constant bitwise-and logand)
(define-constant bitwise-ior logior)
(define-constant bitwise-not lognot)
(define-constant bit-set?    logbit?)

(define-constant fx+ +)
(define-constant fx- -)
(define-constant fx* *)
(define-constant fx/ quotient)
(define-constant fxmod modulo)
(define-constant fxneg -)
(define-constant fxmin min)
(define-constant fxmax max)
(define-constant fx= =)
(define-constant fx> >)
(define-constant fx< <)
(define-constant fx>= >=)
(define-constant fx<= <=)

(define (make-rng init)
  (define (shift n) (fx+ 1 (fx* n 22695477)))
  (let ((s0 (shift (shift (shift (shift init))))))
    (lambda (n)
      (let ((x0 (bitwise-and (arithmetic-shift s0 -16) #b11111111111111)))
        (set! s0 (shift s0))
        (fxmod x0 n)))))

(define getenv sys-getenv)

;; ---END PRELUDE---
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; debug syntax - toggle when we want extra output

(define-syntax debug
  (syntax-rules ()
    ((debug fmt args ...) #t)
    ;;((debug fmt args ...) (format (current-error-port) fmt args ...))
    ))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; R5RS portable terminal library

(define escape (integer->char 27))

(define color-codes
  (let ((tab (make-hash-table)))
    (for-each
     (lambda (c) (hash-table-put! tab (car c) (cdr c)))
     '((clear      .  "0")   (reset      .  "0")
       (bold       .  "1")   (dark       .  "2")
       (underline  .  "4")   (underscore .  "4")
       (blink      .  "5")   (reverse    .  "7")
       (concealed  .  "8")
       (black      . "30")   (on-black   . "40") 
       (red        . "31")   (on-red     . "41") 
       (green      . "32")   (on-green   . "42") 
       (yellow     . "33")   (on-yellow  . "43") 
       (blue       . "34")   (on-blue    . "44") 
       (magenta    . "35")   (on-magenta . "45") 
       (cyan       . "36")   (on-cyan    . "46") 
       (white      . "37")   (on-white   . "47")))
    tab))

(define color
  (let ((base (list->string (list escape #\[))))
    (lambda args
      (string-append
       base (string-join
             (map
              (lambda (x)
                (if (symbol? x)
                  (hash-table-get color-codes x)
                  (hash-table-get color-codes (string->symbol x))))
              args) ";") "m"))))

(define reset (color 'reset))
(define red (color 'red))
(define black (color 'black))
(define green (color 'green))
(define yellow (color 'yellow))
(define white (color 'white))
(define on-red (color 'on-red))
(define on-black (color 'on-black))
(define on-green (color 'on-green))
(define on-yellow (color 'on-yellow))
(define on-white (color 'on-white))

(define home-position (string-append (string escape) "H"))

(define (clear-screen)
  (display (string-append (string escape) "[2J" home-position)))

(define (move-cursor y x)
  (format #t "~A[~A;~AH" (string escape) y x))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; some string utils

(define (string-trim str ch)
  (let ((len (string-length str)))
    (if (and (positive? len)
             (eqv? ch (string-ref str 0)))
      (string-trim (substring str 1 len) ch)
      str)))

(define (string-trim-right str ch)
  (let ((len (- (string-length str) 1)))
    (if (and (not (negative? len))
             (eqv? ch (string-ref str len)))
      (string-trim-right (substring str 0 len) ch)
      str)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; world representation

(define-record-type <cell>
  (make-cell type ant food markers)
  cell?
  (type cell-type set-cell-type!)
  (ant cell-ant set-cell-ant!)
  (food cell-food set-cell-food!)
  (markers cell-markers set-cell-markers!))

(define-record-type <ant>
  (make-ant id state color resting direction has-food?)
  ant?
  (id ant-id set-ant-id!)
  (state ant-state set-ant-state!)
  (color ant-color set-ant-color!)
  (resting ant-resting set-ant-resting!)
  (direction ant-direction set-ant-direction!)
  (has-food? ant-has-food? set-ant-has-food?!))

(define-record-type <world>
  (%make-world table num-cols ants)
  world?
  (table world-table set-world-table!)
  (num-cols world-num-cols set-world-num-cols!)
  (ants world-ants set-world-ants!))

(define-inline (world-num-rows world)
  (quotient (vector-length (world-table world)) (world-num-cols world)))

(define-inline (make-world num-rows num-cols . opt)
  (%make-world (make-vector (* num-rows num-cols) #f) num-cols (get-optional opt #f)))

(define-inline (x+y->pos world x y)
  (+ x (* y (world-num-cols world))))

(define-inline (pos->x+y world pos)
  (receive (y x) (quotient&remainder pos (world-num-cols world))
    (values x y)))

;; can reference cells by either the single position form or by x,y

(define-syntax world-ref
  (syntax-rules ()
    ((world-ref world x y) (vector-ref (world-table world) (x+y->pos world x y) ))
    ((world-ref world pos) (vector-ref (world-table world) pos ))))

(define-syntax world-set!
  (syntax-rules ()
    ((world-set! world x y val)
     (vector-set! (world-table world) (x+y->pos world x y) val))
    ((world-set! world pos val)
     (vector-set! (world-table world) pos val))))

;; optimized adjacent cell forms for fixed-size maps

(define-inline (adjacent-cell-n pos dir n)
  (case dir
    ((0) (fx+ pos 1))
    ((1) (if (even? (fxmod pos n)) (fx+ pos n 1) (fx+ pos n)))
    ((2) (if (even? (fxmod pos n)) (fx+ pos n) (fx+ pos n -1)))
    ((3) (fx- pos 1))
    ((4) (if (even? (fxmod pos n)) (fx- pos n) (fx- pos n 1)))
    ((5) (if (even? (fxmod pos n)) (fx- pos n -1) (fx- pos n)))
    (else (error "invalid direction:" dir))))

(define-inline (adjacent-cell-16 pos dir)
  (case dir
    ((0) (fx+ pos 1))
    ((1) (if (bit-set? 4 pos) (fx+ pos 16) (fx+ pos 17)))
    ((2) (if (bit-set? 4 pos) (fx+ pos 15) (fx+ pos 16)))
    ((3) (fx- pos 1))
    ((4) (if (bit-set? 4 pos) (fx- pos 17) (fx- pos 16)))
    ((5) (if (bit-set? 4 pos) (fx- pos 16) (fx- pos 15)))
    (else (error "invalid direction:" dir))))

(define-inline (adjacent-cell-128 pos dir)
  (case dir
    ((0) (fx+ pos 1))
    ((1) (if (bit-set? 7 pos) (fx+ pos 128) (fx+ pos 129)))
    ((2) (if (bit-set? 7 pos) (fx+ pos 127) (fx+ pos 128)))
    ((3) (fx- pos 1))
    ((4) (if (bit-set? 7 pos) (fx- pos 129) (fx- pos 128)))
    ((5) (if (bit-set? 7 pos) (fx- pos 128) (fx- pos 127)))
    (else (error "invalid direction:" dir))))

;; adjust here, 16 for 10x10 maps and 128 for 100x100 maps, or just
;; always pass n

(define-syntax adjacent-cell
  (syntax-rules ()
    ((adjacent-cell pos dir n) (adjacent-cell-n pos dir n))
    ;;((adjacent-cell pos dir) (adjacent-cell-16 pos dir))
    ((adjacent-cell pos dir) (adjacent-cell-128 pos dir))
    ))

(define-inline (turn side dir)
  (fxmod (+ dir (if (eq? side 'Left) 5 1)) 6))

(define-inline (sensed-cell pos dir sense-dir n)
  (case sense-dir
    ((Here) pos)
    ((Ahead) (adjacent-cell pos dir))
    ((LeftAhead) (adjacent-cell pos (turn 'Left dir)))
    ((RightAhead) (adjacent-cell pos (turn 'Right dir)))
    (else (error "invalid sense-dir:" sense-dir))))

(define-inline (other-color color)
  (if (eq? color 'Red) 'Black 'Red))

(define-inline (set-ant-at! world pos ant)
  (let* ((cell (vector-ref (world-table world) pos)))
    (set-cell-ant! cell ant)))

(define-inline (clear-ant-at! world pos)
  (set-ant-at! world pos #f))

(define-inline (ant-is-alive? world id)
  (vector-ref (world-ants world) id))

(define-inline (kill-ant-at! world pos)
  (debug "~S: Arrrrggghhhhhh!!!\n" (ant-id (cell-ant (world-ref world pos))))
  (vector-set! (world-ants world) (ant-id (cell-ant (world-ref world pos))) #f)
  (clear-ant-at! world pos))

(define-inline (anthill-at? world pos color)
  (eq? color (cell-type (world-ref world x y))))

;; markers

(define-inline (set-marker! cell color i)
  (set-cell-markers!
   cell (bitwise-ior (cell-markers cell)
                     (arithmetic-shift
                      1 (if (eq? color 'Red) i (fx+ i 6))))))

(define-inline (set-marker-at! world pos color i)
  (set-marker! (world-ref world pos) color i))

(define-inline (clear-marker! cell color i)
  (set-cell-markers!
   cell (bitwise-and (cell-markers cell)
                     (bitwise-not (arithmetic-shift
                                   1 (if (eq? color 'Red) i (fx+ i 6)))))))

(define-inline (clear-marker-at! world pos color i)
  (clear-marker! (world-ref world pos) color i))

(define-inline (check-marker? cell color i)
  (bit-set? (if (eq? color 'Red) i (fx+ i 6)) (cell-markers cell)))

(define-inline (check-marker-at? world pos color i)
  (check-marker? (world-ref world pos) color i))

(define-inline (check-any-marker? cell color)
  (let ((markers (cell-markers cell)))
    (if (eq? color 'Red)
      (positive? (bitwise-and markers #b111111))
      (positive? (arithmetic-shift markers -6)))))

(define-inline (check-any-marker-at? world pos color)
  (check-any-marker? (world-ref world pos) color))

(define-inline (cell-matches? cell condition color)
  (if (eq? 'Rocky (cell-type cell))
    (eq? condition 'Rock)
    (case condition
      ((Friend)
       (and-let* ((a (cell-ant cell)))
         (eq? color (ant-color a))))
      ((Foe)
       (and-let* ((a (cell-ant cell)))
         (not (eq? color (ant-color a)))))
      ((FriendWithFood)
       (and-let* ((a (cell-ant cell)))
         (and (eq? color (ant-color a))
              (ant-has-food? a))))
      ((FoeWithFood)
       (and-let* ((a (cell-ant cell)))
         (and (not (eq? color (ant-color a)))
              (ant-has-food? a))))
      ((Food) (positive? (cell-food cell)))
      ((Rock) #f)
      ((Marker0) (check-marker? cell color 0))
      ((Marker1) (check-marker? cell color 1))
      ((Marker2) (check-marker? cell color 2))
      ((Marker3) (check-marker? cell color 3))
      ((Marker4) (check-marker? cell color 4))
      ((Marker5) (check-marker? cell color 5))
      ((FoeMarker) (check-any-marker? cell (other-color color)))
      ((Home) (eq? (cell-type cell) color))
      ((FoeHome) (eq? (cell-type cell) (other-color color)))
      (else (error "invalid condition:" condition)))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; simulator

(define (check-for-surrounded-ant-at world pos)
  (and-let* ((cell (world-ref world pos))
             (a (cell-ant cell)))
    (let loop ((dir 0) (empty 0))
      (if (fx= dir 6)
        (begin
          (kill-ant-at! world pos)
          (set-cell-food! cell (fx+ 3 (cell-food cell))))
        (if (and-let* ((pos (adjacent-cell pos dir))
                       (cell (world-ref world pos))
                       (a2 (cell-ant cell)))
              (not (eq? (ant-color a) (ant-color a2))))
          (loop (fx+ dir 1) empty)
          (let ((x (fx+ empty 1)))
            (unless (fx= x 2) (loop (fx+ dir 1) x))))))))

(define-inline (check-for-surrounded-ants world pos)
  (check-for-surrounded-ant-at world pos)
  (check-for-surrounded-ant-at world (adjacent-cell pos 0))
  (check-for-surrounded-ant-at world (adjacent-cell pos 1))
  (check-for-surrounded-ant-at world (adjacent-cell pos 2))
  (check-for-surrounded-ant-at world (adjacent-cell pos 3))
  (check-for-surrounded-ant-at world (adjacent-cell pos 4))
  (check-for-surrounded-ant-at world (adjacent-cell pos 5)))

(define (step world a pos red black randomint)
  ;;(receive (x y) (pos->x+y world pos)
  ;;  (debug "ant ~S at (~S, ~S): " (ant-id a) (- x 1) (- y 1)))
  (if (positive? (ant-resting a))
    (begin
      (debug "resting: ~S\n" (ant-resting a))
      (set-ant-resting! a (fx- (ant-resting a) 1)))
    (let ((instr (vector-ref (if (eq? 'Red (ant-color a)) red black) (ant-state a))))
      (case (car instr)
        ((Sense) ; sense-dir st1 st2 condition
         (debug "sense: ~S ~S ~S ~S\n" (cadr instr) (caddr instr) (cadddr instr) (fifth instr))
         (let*
             ((sense-pos (sensed-cell pos (ant-direction a) (second instr)
                                      (world-num-cols world)))
              (cell (world-ref world sense-pos)))
           ;;(receive (x y) (pos->x+y world sense-pos)
           ;;  (debug "sensed: (~S, ~S) => ~S\n" (- x 1) (- y 1) (cell-markers cell)))
           (set-ant-state! a (if (cell-matches? cell (fifth instr) (ant-color a)) (third instr) (fourth instr)))))
        ((Mark) ; i st
         (debug "mark: ~S ~S\n" (cadr instr) (caddr instr))
         (set-marker! (world-ref world pos) (ant-color a) (cadr instr))
         (set-ant-state! a (caddr instr)))
        ((Unmark) ; i st
         (debug "unmark: ~S ~S\n" (cadr instr) (caddr instr))
         (clear-marker! (world-ref world pos) (ant-color a) (cadr instr))
         (set-ant-state! a (caddr instr)))
        ((PickUp) ; st1 st2
         (debug "pickup: ~S ~S\n" (cadr instr) (caddr instr))
         (if (ant-has-food? a)
           (set-ant-state! a (caddr instr))
           (let ((cell (world-ref world pos)))
             (if (zero? (cell-food cell))
               (set-ant-state! a (caddr instr))
               (begin
                 (set-cell-food! cell (fx- (cell-food cell) 1))
                 (set-ant-has-food?! a #t)
                 (set-ant-state! a (cadr instr)))))))
        ((Drop) ; st
         (debug "drop: ~S\n" (cadr instr))
         (when (ant-has-food? a)
           (let ((cell (world-ref world pos)))
             (set-cell-food! cell (fx+ 1 (cell-food cell)))
             (set-ant-has-food?! a #f)))
         (set-ant-state! a (cadr instr)))
        ((Turn) ; lr st
         (debug "turn: ~S ~S\n" (cadr instr) (caddr instr))
         (set-ant-direction! a (turn (cadr instr) (ant-direction a)))
         (set-ant-state! a (caddr instr)))
        ((Move) ; st1 st2
         (debug "move: ~S ~S\n" (cadr instr) (caddr instr))
         (let* ((new-pos (adjacent-cell pos (ant-direction a))) ; (world-num-cols world)
                (new-cell (world-ref world new-pos)))
           ;;(receive (x y) (pos->x+y world new-pos)
           ;;  (debug "new-pos: ~S (~S, ~S), new-cell: ~S (~S)\n"
           ;;         new-pos (- x 1) (- y 1) new-cell (and-let* ((a (cell-ant new-cell))) (ant-id a))))
           (if (or (eq? 'Rocky (cell-type new-cell)) (cell-ant new-cell))
             (set-ant-state! a (caddr instr))
             (begin
               (clear-ant-at! world pos)
               (set-cell-ant! new-cell a)
               (set-ant-state! a (cadr instr))
               (set-ant-resting! a 14)
               (set-cdr! (vector-ref (world-ants world) (ant-id a)) new-pos)
               (check-for-surrounded-ants world new-pos)))))
        ((Flip) ; p st1 st2
         (debug "flip: ~S ~S ~S\n" (cadr instr) (caddr instr) (cadddr instr))
         (set-ant-state! a (if (zero? (randomint (cadr instr))) (caddr instr) (cadddr instr))))
        (else (warn "invalid instruction: ~S" instr))))))

(define (run world red black . opt)
  (let-optionals* opt
      ((seed 12345)
       (num-steps *default-run-steps*)
       (round-callback #f)
       (step-callback #f))
    (let* ((ants (world-ants world))
           (num-ants (vector-length ants))
           (randomint (if (procedure? seed) seed (make-rng seed))))
      (cond
        (step-callback
         (dotimes (i num-steps)
           (dotimes (j num-ants)
             (and-let* ((a+p (vector-ref ants j)))
               ;;(step world (car a+p) (cdr a+p) red black randomint)
               (step-callback step world (car a+p) (cdr a+p) red black randomint)))
           (if round-callback (round-callback world i))))
        (round-callback
         (dotimes (i num-steps)
           (dotimes (j num-ants)
             (and-let* ((a+p (vector-ref ants j)))
               (step world (car a+p) (cdr a+p) red black randomint)))
           (round-callback world i)))
        (else
         (dotimes (i num-steps)
           (dotimes (j num-ants)
             (and-let* ((a+p (vector-ref ants j)))
               (step world (car a+p) (cdr a+p) red black randomint)))))))))

(define (score world)
  (let* ((table (world-table world))
         (len (vector-length table))
         (red-score 0)
         (black-score 0))
    (do ((i 0 (fx+ i 1)))
        ((= i len))
      (and-let* ((cell (vector-ref table i)))
        (case (cell-type cell)
          ((Red) (inc! red-score (cell-food cell)))
          ((Black) (inc! black-score (cell-food cell))))))
    (list red-score black-score)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; I/O utils

(define (parse-instruction line)
  (let ((ls (call-with-input-string line (lambda (p) (port->list read p)))))
    (when (and (eq? (car ls) 'Sense) (eq? (car (cddddr ls)) 'Marker))
      (set-car! (cddddr ls) (case (cadr (cddddr ls))
                              ((0) 'Marker0) ((1) 'Marker1) ((2) 'Marker2)
                              ((3) 'Marker3) ((4) 'Marker4) ((5) 'Marker5)))
      (set-cdr! (cddddr ls) '()))
    ls))

(define (load-state-machine file)
  (call-with-input-file file
    (lambda (p)
      (list->vector (map parse-instruction (port->list read-line p))))))

(define (load-world file)
  (define (next-power-of-two n) ; just handle likely map sizes
    (cond ((< n 16) 16)
          ((< n 128) 128)
          ((< n 1024) 1024)))
  (define (char->cell-type c)
    (case c
      ((#\+) 'Red)    ((#\-) 'Black)
      ((#\#) 'Rocky)  (else  'Clear)))
  (define (char->food c)
    (let ((i (char->integer c)))
      (if (<= 48 i 57) (- i 48) 0)))
  (define (make-counter n)
    (lambda () (let ((res n)) (set! n (+ n 1)) res)))
  (with-input-from-file file
    (lambda ()
      (let* ((num-cols (string->number (read-line)))
             (n (next-power-of-two num-cols))
             (num-rows (string->number (read-line)))
             (m (+ num-rows 2))
             (world (make-world m n 0))
             (ant-id 0)
             (ants '()))
        (define (next-ant color)
          (let ((res (make-ant ant-id 0 color 0 0 #f)))
            (inc! ant-id)
            res))
        (let loop ((y 1))
          (let ((line (read-line)))
            (cond
              ((eof-object? line)
               (unless (= y (- m 1)) (warn "premature end of world")))
              ((> y num-rows)
               (warn "spurious extra lines: ~S" line))
              (else
               (let ((x 1))
                 (for-each
                  (lambda (c)
                    (let* ((type (char->cell-type c))
                           (ant (case type ((Red Black) (next-ant type)) (else #f)))
                           (food (char->food c)))
                      (world-set! world x y (make-cell type ant food 0))
                      (when ant
                        (push! ants (cons ant (x+y->pos world x y))))
                      (inc! x)))
                  (filter (lambda (c) (not (eqv? c #\space))) (string->list line))))
               (loop (+ y 1))))))
        (set-world-ants! world (list->vector (reverse ants)))
        world))))

(define (print-world world . opt)
  (let* ((ansi? (get-optional opt #f))
         (table (world-table world))
         (num-cols (world-num-cols world))
         (num-rows (world-num-rows world)))
    (if ansi?
      (let ((prev-fg #f) (prev-bg #f))
        (do ((y 1 (+ y 1)))
            ((= y (- num-rows 1)))
          (if (even? y) (display " "))
          (do ((x 1 (+ x 1)))
              ((= x (- num-cols 1)))
            (and-let* ((cell (world-ref world x y))
                       (food (min 10 (or (cell-food cell) 0)))
                       (bg (case (cell-type cell)
                             ((Rocky) on-yellow)
                             ((Red) on-red)
                             ((Black) on-black)
                             (else (if (zero? food) on-white on-green)))))
              (let ((fg white)
                    (ch " "))
                (cond ((cell-ant cell)
                       => (lambda (a)
                            (cond ((eq? (ant-color a) 'Red)
                                   (set! ch "r")
                                   (unless (eq? bg on-red) (set! fg red)))
                                  (else
                                   (set! ch "b")
                                   (unless (eq? bg on-black) (set! fg black))))))
                      ((not (zero? food))
                       (set! ch (if (= 10 food) "0" (number->string food)))
                       (unless (eq? bg on-green) (set! fg green))))
                (unless (eq? fg prev-fg) (display fg))
                (unless (eq? bg prev-bg) (display bg))
                (display ch) (display " ")
                (set! prev-fg fg)
                (set! prev-bg bg))))
          (display reset)
          (set! prev-fg #f)
          (set! prev-bg #f)
          (newline)))
      (do ((y 1 (+ y 1)))
          ((= y (- num-rows 1)))
        (if (even? y) (display " "))
        (do ((x 1 (+ x 1)))
            ((= x (- num-cols 1)))
          (and-let* ((cell (world-ref world x y)))
            (case (cell-type cell)
              ((Rocky) (display "#"))
              ((Red) (display "+"))
              ((Black) (display "-"))
              (else
               (if (positive? (cell-food cell))
                 (display (min 9 (cell-food cell)))
                 (display "."))))
            (display " ")))
        (newline)))))

(define (paint-cell world pos)
  (receive (x y) (pos->x+y world pos)
    (and-let* ((cell (world-ref world pos))
               (food (min 10 (or (cell-food cell) 0)))
               (bg (case (cell-type cell)
                     ((Rocky) on-yellow)
                     ((Red) on-red)
                     ((Black) on-black)
                     (else (if (zero? food) on-white on-green)))))
      (let ((fg white)
            (ch " "))
        (cond ((cell-ant cell)
               => (lambda (a)
                    (cond ((eq? (ant-color a) 'Red)
                           (set! ch "r")
                           (unless (eq? bg on-red) (set! fg red)))
                          (else
                           (set! ch "b")
                           (unless (eq? bg on-black) (set! fg black))))))
              ((not (zero? food))
               (set! ch (if (= 10 food) "0" (number->string food)))
               (unless (eq? bg on-green) (set! fg green))))
        (move-cursor y (+ (* x 2) (if (even? y) 1 0)))
        (display fg) (display bg) (display ch) (display " ") (display reset)))))

;; graphical stepper

(define (step-world step world a pos red black randomint)
  (let ((res (step world a pos red black randomint)))
    (and-let* ((a+p (vector-ref (world-ants world) (ant-id a)))
               (new-pos (cdr a+p))
               ((not (= pos new-pos))))
      (paint-cell world pos)
      (paint-cell world new-pos))
    res))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; for verification

(define (markers->string n)
  (string-join (map (lambda (i) (if (logbit? i n) (number->string i) "")) (iota 6)) ""))

(define (ant->string a)
  (format "~A ant of id ~A, dir ~A, food ~A, state ~A, resting ~A"
          (if (eq? 'Red (ant-color a)) "red" "black")
          (ant-id a) (ant-direction a) (if (ant-has-food? a) 1 0)
          (ant-state a) (ant-resting a)))

(define (cell->string cell)
  (let ((ls (case (cell-type cell)
              ((Rocky) '("rock"))
              ((Red) '("red hill"))
              ((Black) '("black hill"))
              (else '()))))
    (if (positive? (cell-food cell)) (set! ls (append ls (list (format "~D food" (cell-food cell))))))
    (let* ((marks (cell-markers cell))
           (red-marks (logand marks #b111111))
           (black-marks (logand (ash marks -6) #b111111)))
      (if (positive? red-marks) (push! ls (format "red marks: ~A" (markers->string red-marks))))
      (if (positive? black-marks) (push! ls (format "black marks: ~A" (markers->string black-marks)))))
    (if (cell-ant cell) (push! ls (ant->string (cell-ant cell))))
    (string-join (reverse ls) "; ")))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; main

(define-constant *default-run-steps* 100000)

(define (main args)
  (let ((round-callback #f)
        (step-callback #f)
        (callback-init #f)
        (seed 12345)
        (steps *default-run-steps*)
        (verbose? #f)
        (rate 1000)
        (ansi? (and-let* ((term (getenv "TERM")))
                 (member term '("vt100" "linux" "xterm" "xterm-color" "mlterm")))))
    (define (add-round-callback proc)
      (if round-callback
        (let ((orig round-callback))
          (set! round-callback (lambda (w i) (orig w i) (proc w i))))
        (set! round-callback proc)))
    (define (show world i)
      (when (zero? (fxmod i rate))
        (newline)
        (format #t "Step: ~S\n" i)
        (print-world world ansi?)))
    (define (go world-file red-file black-file)
      (let ((world (load-world world-file))
            (red (load-state-machine red-file))
            (black (load-state-machine black-file)))
        (if verbose? (debug "red: ~S\n" red))
        (if verbose? (debug "ants: ~S (~S)\n" (world-ants world) (vector-length (world-ants world))))
        (if callback-init (callback-init world -1))
        (run world red black seed steps round-callback step-callback)
        (print (score world))))
    (let loop ((ls (cdr args)))
      (if (eqv? (string-ref (car ls) 0) #\-)
        (case (string->symbol (car ls))
          ((-c -check)
           (let ((p (open-input-file (cadr ls))))
             (define (verify w i)
               (read-line p)
               (let ((str1 (format "After round ~D..." (+ 1 i)))
                     (str2 (read-line p)))
                 (unless (string=? str1 str2) (warn "invalid round #: ~S != ~S" str1 str2)))
               (let ((num-rows (world-num-rows w))
                     (num-cols (world-num-cols w)))
                 (dotimes (y num-rows)
                   (dotimes (x num-cols)
                     (and-let*
                         ((cell (world-ref w x y))
                          (line (read-line p))
                          (ls (string-split line "): "))
                          ((pair? (cdr ls)))
                          (str1 (cadr ls))
                          (str2 (cell->string cell)))
                       (unless (string-ci=? (string-trim-right (string-trim-right str1 #\space) #\;) str2)
                         (warn "mismatch in round ~D at (~A, ~A): got ~S, expected ~S" i (- x 1) (- y 1) str2 str1)))))))
             (read-line p)
             (add-round-callback verify)
             (set! callback-init verify)
             (loop (cddr ls))))
          ((-a -ansi) (set! ansi? #t) (loop (cdr ls)))
          ((-g -gui) (add-round-callback show) (loop (cdr ls)))
          ((-w -world) (print-world (load-world (cadr ls)) ansi?) (exit 0))
          ((-p -steps) (set! steps (string->number (cadr ls))) (loop (cddr ls)))
          ((-r -rate) (set! rate (string->number (cadr ls))) (loop (cddr ls)))
          ((-s -seed) (set! seed (string->number (cadr ls))) (loop (cddr ls)))
          ((-t -trace)
           (set! ansi? #t)
           (set! step-callback step-world)
           (set! callback-init (lambda (w i) (clear-screen) (print-world w i)))
           (loop (cdr ls)))
          ((-v -verbose) (set! verbose? #t) (loop (cdr ls)))
          (else (error "unknown option:" (car ls))))
        (case (length ls)
          ((2) (apply go (append ls (cdr ls))))
          ((3) (apply go ls))
          (else (format #t "usage: ant [options] world red-ant [black-ant]\n")))))))

