#!/usr/local/bin/gosh

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; ---START GAUCHE PRELUDE---

(use srfi-1)
(use srfi-2)
(use srfi-9)
(use util.list)
(use file.util)

(define-constant arithmetic-shift ash)
(define-constant bitwise-and logand)
(define-constant bitwise-ior logior)
(define-constant bitwise-not lognot)
(define-constant bit-set?    logbit?)

;; ---END PRELUDE---
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (cond->if clauses)
  (if (null? clauses)
    '()
    (let ((this (car clauses)))
      (if (eq? 'else (car this))
        (cdr this)
        `((if ,(car this)
            (begin ,@(cdr this))
            ,@(cond->if (cdr clauses))))))))

(define (case->if ls)
  (let ((var (car ls)))
    (let loop ((clauses (cdr ls)))
      (if (null? clauses)
        '()
        (let ((this (car clauses)))
          (cond
            ((eq? 'else (car this))
             (cdr this))
            ((= 1 (length (car this)))
             `((if (equal? ,var , (caar this))
                 (begin ,@(cdr this))
                 ,@(loop (cdr clauses)))))
            (else
             `((if (or ,@(map (lambda (x) `(equal? ,var ,x)) (car this)))
                 (begin ,@(cdr this))
                 ,@(loop (cdr clauses)))))))))))

(define (list-pad ls i . opt)
  (let ((diff (- i (length ls))))
    (if (positive? diff)
      (append ls (apply make-list diff opt))
      ls)))

(define (label->symbol x)
  (let ((str (if (keyword? x) (keyword->string x) (symbol->string x))))
    (string->symbol (regexp-replace #/^\:+/ (regexp-replace #/\:+$/ str "") ""))))

(define (replace* orig from to)
  (let loop ((x orig))
    (cond ((eq? x from) to)
          ((pair? x) (cons (loop (car x)) (loop (cdr x))))
          (else x))))

(define (color-labels body)
  (let* ((labels (filter symbol? body))
         (colored (map (lambda (l) (gensym (symbol->string l))) labels)))
    (let loop ((res body) (l labels) (c colored))
      (if (null? l)
        res
        (loop (replace* (replace* res (car l) (car c)) (label->symbol (car l)) (car c))
              (cdr l) (cdr c))))))

(define (param-subst orig params args)
  ;;(warn "param-subst: ~S ~S ~S" orig params args)
  (cond
    ((pair? params)
     (param-subst (replace* orig (car params) (car args)) (cdr params) (cdr args)))
    ((null? params) orig)
    (else (replace* orig params args))))

(define env-set! hash-table-put!)
(define env-ref hash-table-get)

(define-syntax make-default-env
  (syntax-rules ()
    ((make-env sym ...)
     (let ((env (make-hash-table)))
       (for-each (cut env-set! env <> <>) '(sym ...) (list sym ...))
       env))))

(define (~or . args) (any identity args))
(define (~and . args) (every identity args))
(define ~not not)
(define (true) #t)
(define (false) #f)

(define (make-env)
  (make-default-env
   + - * / = < <= > >= max min abs quotient remainder modulo
   ~or ~and ~not gcd lcm zero? even? odd? true false eq? eqv? equal?
   arithmetic-shift bitwise-and bitwise-ior bitwise-not bit-set?))

(define undef (undefined))

(define (compile-internal src . opt)
  (let-optionals* opt
      ((env (make-env))
       (state 0)
       (cont #f))
    (let ((code '()))
      (let loop ((ls src))
        (define (build-test code state pass-label fail-label)
          (case (car code)
            ((Move PickUp)
             (if (= 1 (length code))
               `((,@code ,pass-label ,fail-label))
               (error "wrong number of arguments: ~S" code)))
            ((Flip)
             (if (= 2 (length code))
               `((,@code ,pass-label ,fail-label))
               (error "wrong number of arguments to Flip")))
            ((Here Ahead LeftAhead RightAhead)
             `((Sense ,@code ,pass-label ,fail-label)))
            ((and or)
             (let* ((tmp-label (gensym))
                    (a (if (eq? 'and (car code))
                         (build-test (cadr code) state tmp-label fail-label)
                         (build-test (cadr code) state pass-label tmp-label)))
                    (state2 (+ state (length a)))
                    (b (if (null? (cdddr code))
                         (build-test (caddr code) state2 pass-label fail-label)
                         (build-test `(,(car code) ,@(cddr code)) state2 pass-label fail-label))))
               (env-set! env tmp-label state2)
               `(,@a ,@b)))
            ((not)
             (build-test (cadr code) state fail-label pass-label))
            (else
             (let ((op (ant-eval (car code))))
               (if (procedure? op)
                 (not (not (apply op (map ant-eval (cdr code)))))
                 (errorf "unknown test form: ~S = ~S" code op))))))
        (define (ant-eval expr)
          ;;(warn "ant-eval: ~S" expr)
          (cond
            ((number? expr) expr)
            ((promise? expr) (force expr))
            ((memq expr '(Here Left Right Ahead LeftAhead RightAhead Food Home
                          Red Black Friend Foe FriendWithFood FoeWithFood Food
                          Rock Marker FoeMarker Home FoeHome))
             expr)
            ((eq? expr '$) (if (and (null? (cdr ls)) cont) (ant-eval cont) (+ state 1)))
            ((pair? expr)
             (let ((op (ant-eval (car expr))))
               (cond
                 ((procedure? op)
                  (apply op (map force (map ant-eval (cdr expr)))))
                 ((and (pair? op) (eq? (car op) 'fun))
                  (ant-eval (car (param-subst (cddr op) (cdadr op) (cdr expr)))))
                 (else
                  (case (car expr)
                    ((if)
                     (if (ant-eval (cadr expr))
                       (ant-eval (caddr expr))
                       (and (pair? (cdddr expr)) (ant-eval (cadddr expr)))))
                    ((cond) (ant-eval (car (cond->if (cdr expr)))))
                    ((case) (ant-eval (car (case->if (cdr expr)))))
                    ((begin) (last (map ant-eval (cdr expr))))
                    (else (error "unknown operator:" expr)))))))
            (else
             (let* ((sym expr)
                    (val (env-ref env sym undef)))
               (if (undefined? val)
                 (delay (let ((res (env-ref env sym #f)))
                                (unless res
                                  (warn "unbound symbol: ~S (~S) in ~S" sym res env))
                                res))
                 val)))))
        (define (add-code op args)
          (push! code (cons op (map ant-eval args)))
          (inc! state))
        ;;(warn "loop: ~S" (and (pair? ls) (car ls)))
        (if (null? ls)
          (reverse code)
          (let ((expr (car ls)))
            (cond
              ((or (symbol? expr) (keyword? expr))
               (env-set! env (label->symbol expr) state)
               (loop (cdr ls)))
              ((pair? expr)
               (let ((op (car expr)) (args (cdr expr)))
                 (case op
                   ((define)
                    (if (pair? (car args))
                      (env-set! env (caar args) (cons 'fun args))
                      (env-set! env (car args) (ant-eval (cadr args))))
                    (loop (cdr ls)))
                   ((defsub)
                    (warn "non-top-level defsub: ~S" expr)
                    (env-set! env (caar args) (cons 'sub args)))
                   ((begin) (loop (append args (cdr ls))))
                   ((if)
                    (let* ((pass-label (gensym))
                           (fail-label (gensym))
                           (end-label (gensym))
                           (test-code (build-test (car args) state pass-label fail-label)))
                      (cond
                        ((boolean? test-code)
                         (loop (append (if test-code (list (cadr args)) (cddr args)) (cdr ls))))
                        (else
                         (let ((test (compile-internal test-code env state)))
                           (set! code (append (reverse test) code))
                           (inc! state (length test))
                           (env-set! env pass-label state)
                           (let* ((pass-src (list (cadr args)))
                                  (pass (compile-internal pass-src env state end-label)))
                             (set! code (append (reverse pass) code))
                             (inc! state (length pass))
                             (env-set! env fail-label state)
                             (when (pair? (cddr args))
                               (let* ((fail-src (cddr args))
                                      (fail (compile-internal fail-src env state)))
                                 (set! code (append (reverse fail) code))
                                 (inc! state (length fail))))
                             (env-set! env end-label state)
                             (loop (cdr ls))))))))
                   ((cond) (loop (append (cond->if (cdr expr)) (cdr ls))))
                   ((case) (loop (append (case->if (cdr expr)) (cdr ls))))
                   ((Sense)
                    (let ((n (if (eq? 'Marker (cadr args)) 2 1))
                          (rest (cdr args)))
                      (add-code op (cons (car args) (append (list-pad (drop rest n) 2 '$) (take rest n))))
                      (loop (cdr ls))))
                   ((Drop)
                    (add-code op (list-pad args 1 '$))
                    (loop (cdr ls)))
                   ((Mark Unmark PickUp Turn Move)
                    (add-code op (list-pad args 2 '$))
                    (loop (cdr ls)))
                   ((Flip) (add-code op (list-pad args 3 '$)) (loop (cdr ls)))
                   ((State) (push! code (car args)) (inc! state) (loop (cdr ls)))
                   (else
                    (let ((sub (env-ref env op #f)))
                      (define (state? x)
                        (and (pair? x) (pair? (car x)) (eq? 'State (caar x))))
                      (cond
                        ((and (pair? sub) (eq? (car sub) 'sub))
                         (let* ((params (cdadr sub))
                                (src (if (state? sub)
                                       (cddr sub)
                                       (color-labels
                                        (param-subst (cddr sub) params args))))
                                (end-label (gensym)))
                           (env-set! env state (if (state? sub)
                                                 (format "  (~A)" op)
                                                 (format "~A" op)))
                           (env-set! env op `(sub (State) (State ,state)))
                           (let ((res (compile-internal src env state end-label)))
                             (env-set! env op sub)
                             (set! code (append (reverse res) code))
                             (inc! state (length res))
                             (env-set! env end-label state)
                             (loop (cdr ls)))))
                        (else
                         (errorf "unhandled expr: ~S => ~S" expr sub))))))))
              (else (error "unhandled expr:" expr)))))))))

(define (compile src . args)
  (let ((env (get-optional args (make-env))))
    (receive (subs code)
        (partition (lambda (x) (and (pair? x) (eq? 'defsub (car x)))) src)
      (for-each
       (lambda (x) (env-set! env (caadr x) (cons 'sub (cdr x))))
       subs)
      (let* ((vec (list->vector
                   (apply compile-internal code env
                          (if (pair? args) (cdr args) '()))))
             (len (vector-length vec)))
        (if (>= len 10000)
          (error "program too large"))
        (do ((i 0 (+ i 1)))
            ((= i len))
          (if (number? (vector-ref vec i))
            (vector-set! vec i (vector-ref vec (vector-ref vec i)))))
        (map (lambda (x) (if (pair? x) (map force x) x))
             (vector->list vec))))))

(define (instr->string instr)
  (with-output-to-string
    (lambda ()
      (display (car instr))
      (for-each (lambda (x) (format #t " ~A" x)) (cdr instr)))))

(define (read-file file)
  (call-with-input-file file
    (lambda (p) (port->list read p))))

(define (preprocess ls)
  (define (find-file file)
    (if (file-exists? file)
      file
      (find file-exists?
            (map (cut build-path <> file)
                 (string-split (or (sys-getenv "ANT_PATH") "") #/:/)))))
  (let loop ((ls ls) (res '()))
    (cond
      ((null? ls)
       (reverse res))
      ((and (pair? (car ls)) (eq? 'include (caar ls)))
       (let ((file (find-file (cadar ls))))
         (if file
           (loop (append (read-file file) (cdr ls)) res)
           (error "couldn't find file: " file))))
      (else (loop (cdr ls) (cons (car ls) res))))))

(define (main args)
  (let ((verbose? #f)
        (comments? #f))
    (let loop ((ls (cdr args)))
      (if (eqv? (string-ref (car ls) 0) #\-)
        (case (string->symbol (car ls))
          ((-v -verbose) (set! verbose? #t) (loop (cdr ls)))
          ((-c -comments) (set! comments? #t) (loop (cdr ls)))
          (else (error "unknown option:" (car ls))))
        (let* ((env (make-env))
               (src (preprocess (read-file (car ls))))
               (code (compile src env)))
          (if comments?
            (let output ((ls code) (i 0))
              (when (pair? ls)
                (cond
                  ((env-ref env i #f)
                   => (lambda (str)
                        (format #t "~32A ; ~3D: ~A\n" (instr->string (car ls)) i str)))
                  (else
                   (format #t "~32A ; ~3D:\n" (instr->string (car ls)) i)))
                (output (cdr ls) (+ i 1))))
            (map print (map instr->string code)))
          0)))))
